/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Pierre Garcia
 * Contributor(s): 
 */

package org.objectweb.dream.dreamannotation.processor;

import java.lang.annotation.Annotation;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;

import org.objectweb.dream.control.lifecycle.PrepareStopLifeCycleController;
import org.objectweb.dream.dreamannotation.DreamComponent;
import org.objectweb.dream.dreamannotation.DreamLifeCycle;
import org.objectweb.dream.dreamannotation.generator.DreamComponentADLGenerator;
import org.objectweb.dream.dreamannotation.generator.DreamDefinitionGenerator;
import org.objectweb.dream.dreamannotation.generator.template.binding.DreamRequiresTemplate;
import org.objectweb.dream.dreamannotation.generator.template.lifecycle.DreamLifeCycleAsyncStartTemplate;
import org.objectweb.dream.dreamannotation.generator.template.lifecycle.DreamLifeCyclePrepareStopTemplate;
import org.objectweb.dream.dreamannotation.generator.template.lifecycle.DreamLifeCycleTemplate;
import org.objectweb.dream.dreamannotation.generator.template.lifecycle.DreamLifeCycleTemplateFirstStart;
import org.objectweb.dream.dreamannotation.generator.template.lifecycle.DreamLifeCycleTemplateFirstStartAndStart;
import org.objectweb.dream.dreamannotation.generator.template.lifecycle.DreamLifeCycleTemplateFirstStartAndStop;
import org.objectweb.dream.dreamannotation.generator.template.lifecycle.LifeCycleTemplate;
import org.objectweb.dream.dreamannotation.generator.template.lifecycle.LifeCycleTemplateStart;
import org.objectweb.dream.dreamannotation.generator.template.lifecycle.LifeCycleTemplateStop;
import org.objectweb.dream.dreamannotation.util.DreamLifeCycleType;
import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.api.control.LifeCycleController;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.fractal.fraclet.annotation.annotations.Service;
import org.objectweb.fractal.fraclet.annotation.generator.ComponentADLGenerator;
import org.objectweb.fractal.fraclet.annotation.generator.Generator;
import org.objectweb.fractal.fraclet.annotation.generator.component.AttributeGenerator;
import org.objectweb.fractal.fraclet.annotation.generator.component.ContentGenerator;
import org.objectweb.fractal.fraclet.annotation.generator.component.LicenceGenerator;
import org.objectweb.fractal.fraclet.annotation.generator.component.ProvidesGenerator;
import org.objectweb.fractal.fraclet.annotation.generator.component.RequiresGenerator;
import org.objectweb.fractal.fraclet.annotation.generator.template.util.NonTypedFractalAttributeException;

import spoon.processing.AbstractProcessor;
import spoon.processing.Property;
import spoon.reflect.declaration.CtAnnotation;
import spoon.reflect.declaration.CtClass;
import spoon.reflect.declaration.CtElement;
import spoon.reflect.declaration.CtField;
import spoon.reflect.declaration.CtMethod;
import spoon.reflect.reference.CtTypeReference;
import spoon.template.Substitution;

/**
 * An annotation processor to manage the FractalComponent and the DreamComponent
 * annotation (generation of ".fractal" file).
 * 
 * @author Pierre Garcia <pierre.garcia@inrialpes.fr>
 */
public class DreamComponentProcessor extends AbstractProcessor<CtClass<?>> {
	private static final boolean DEBUG = false;
	
	@Property
	Class<?>[] generatorClass = new Class[] { LicenceGenerator.class,
			DreamDefinitionGenerator.class, RequiresGenerator.class,
			ProvidesGenerator.class, ContentGenerator.class,
			AttributeGenerator.class };

	private List<Generator> generators = new ArrayList<Generator>();

	@Override
	public void init() {
		super.init();

		generators.clear();

		for (Class<?> c : generatorClass) {
			try {
				generators.add((Generator) c.newInstance());
			} catch (InstantiationException e) {
				e.printStackTrace();
			} catch (IllegalAccessException e) {
				e.printStackTrace();
			}
		}

	}

	/**
	 * A constructor for the DreamComponentProcessor
	 * 
	 * @param processedClass
	 *            the CtClass processed
	 */
	public void process(CtClass<?> processedClass) {
		// field annotated with @Service
		CtField<?> weaveable = null;

		FractalComponent fcAnnotation = processedClass
				.getAnnotation(FractalComponent.class);
		if (fcAnnotation != null) {
			System.out.println("[INFO] Processing "
					+ processedClass.getQualifiedName()
					+ " fractal component");
			new ComponentADLGenerator(processedClass, fcAnnotation, generators);
			try {
				Substitution
						.insertAll(processedClass, new DreamRequiresTemplate(
								getFactory(), processedClass));
				weaveable = getServiceField(processedClass);
				if (weaveable != null) {
					weaveable.getAnnotations().remove(
							weaveable.getAnnotation(getFactory().Type()
									.createReference(Service.class)));
				}
				new DreamRequiresTemplate(getFactory(), processedClass);
			} catch (NonTypedFractalAttributeException e) {
				e.printStackTrace();
			}
		} else {
			// Test if it is a dream component
			DreamComponent dcAnnotation = processedClass
					.getAnnotation(DreamComponent.class);

			if (dcAnnotation != null) {
				System.out.println("[INFO] Processing "
						+ processedClass.getQualifiedName()
						+ " dream component");
				new DreamComponentADLGenerator(processedClass, dcAnnotation,
						generators);
				try {
					weaveable = getServiceField(processedClass);
					Substitution.insertAll(processedClass,
							new DreamRequiresTemplate(getFactory(),
									processedClass));
					weaveable = getServiceField(processedClass);
					CtAnnotation<Service> ctas = null;
					if (weaveable != null) {
						if (DEBUG) System.out.print("REMOVING service annotation in "
								+ processedClass.getQualifiedName()
								+ " dream component \n");
						ctas = weaveable.getAnnotation(getFactory().Type()
								.createReference(Service.class));
						weaveable.getAnnotations().remove(ctas);
					}
					new DreamRequiresTemplate(getFactory(), processedClass);
					if (weaveable != null) {
						if (DEBUG) System.out.print("ADDING service annotation in "
								+ processedClass.getQualifiedName()
								+ " dream component \n");
						Set<CtAnnotation<? extends Annotation>> annotationSet = weaveable
								.getAnnotations();
						annotationSet = weaveable.getAnnotations();
						annotationSet.add(ctas);
						weaveable.setAnnotations(annotationSet);
					}
				} catch (NonTypedFractalAttributeException e) {
					e.printStackTrace();
				}
				// management of the needAsyncStart parameter
				Substitution.insertAll(processedClass,
						new DreamLifeCycleAsyncStartTemplate(dcAnnotation
								.needAsyncStart()));
			}
		}

		// Manage LifeCycle annotation
		List<CtElement> elts = processedClass
				.getAnnotatedChildren(LifeCycle.class);
		boolean start = false, stop = false, done = false;
		CtMethod<?> startM = null, stopM = null;
		CtClass<?> currentProcessedClass = processedClass;

		while (currentProcessedClass != null) {
			elts = currentProcessedClass.getAnnotatedChildren(LifeCycle.class);
			for (CtElement elt : elts) {
				LifeCycle annot = elt.getAnnotation(LifeCycle.class);

				if (annot.on().equals(LifeCycleType.START) && (!start)) {
					start = true;
					startM = (CtMethod<?>) elt;
					if (startM.getParameters().size() > 0)
						System.err
								.println("WARNING: a method annotated with @LifeCycle shouldn't "
										+ "have parameters! Please revise your method signature.");
					if (stop) {
						done = true;
						break;
					}
				}

				if (annot.on().equals(LifeCycleType.STOP) && (!stop)) {
					stop = true;
					stopM = (CtMethod<?>) elt;
					if (stopM.getParameters().size() > 0)
						System.err
								.println("WARNING: a method annotated with @LifeCycle shouldn't "
										+ "have parameters! Please revise your method signature.");
					if (start) {
						done = true;
						break;
					}
				}
			}
			if (done)
				break;
			if (currentProcessedClass.getReference().getSuperclass() != null) {
				currentProcessedClass = (CtClass<?>) currentProcessedClass
						.getReference().getSuperclass().getDeclaration();
			} else {
				currentProcessedClass = null;
			}
		}

		// Manage DreamLifeCycle annotation
		boolean dreamFirstStart = false, dreamPrepareStop = false;
		CtMethod<?> dreamFirstStartM = null, dreamPrepareStopM = null;
		elts = processedClass.getAnnotatedChildren(DreamLifeCycle.class);

		currentProcessedClass = processedClass;
		done = false;

		while (currentProcessedClass != null) {
			elts = currentProcessedClass.getAnnotatedChildren(DreamLifeCycle.class);
			for (CtElement elt : elts) {
				DreamLifeCycle annot = elt.getAnnotation(DreamLifeCycle.class);

				if (annot.on().equals(DreamLifeCycleType.FIRST_START)
						&& (!dreamFirstStart)) {
					// check the existance of @Service annotations
					if (weaveable != null) {
						dreamFirstStart = true;
						dreamFirstStartM = (CtMethod<?>) elt;

						if (dreamFirstStartM.getParameters().size() != 1) {
							System.err
									.println("\n### ERROR ###\nA method annotated with @DreamLifeCycle(on=@DreamLifeCycleType.FIRST_START) should "
											+ "have exactly one parameter of type Component! Please revise your method signature.\n#############\n");
						}
						if (dreamPrepareStop) {
							done = true;
							break;
						}
					} else {
						System.err
								.println("\n### ERROR ###\nThere is no @Service annotation declared in "
										+ processedClass.getQualifiedName()
										+ " \n"
										+ "You can't use @DreamLifeCycle(on=DreamLifeCycleType.First_Start) without a component service declared \n#############\n");
					}
				}

				if (annot.on().equals(DreamLifeCycleType.PREPARE_TO_STOP)
						&& (!dreamPrepareStop)) {
					dreamPrepareStop = true;
					dreamPrepareStopM = (CtMethod<?>) elt;
					if (dreamPrepareStopM.getParameters().size() > 0)
						System.err
								.println("\n### ERROR ###\nA method annotated with @LifeCycle must not "
										+ "have parameters! Please revise your method signature.\n#############\n");
					if (dreamFirstStart) {
						done = true;
						break;
					}
				}
			}
			if (done)
				break;
			if (currentProcessedClass.getReference().getSuperclass() != null) {
				currentProcessedClass = (CtClass<?>) currentProcessedClass
						.getReference().getSuperclass().getDeclaration();
			} else {
				currentProcessedClass = null;
			}
		}

		if (stop) {
			if (start) {
				if (dreamFirstStart) {
					Substitution.insertAll(processedClass,
							new DreamLifeCycleTemplate(startM, stopM,
									dreamFirstStartM, weaveable, processedClass.getActualClass()));
				} else {
					Substitution.insertAll(processedClass,
							new LifeCycleTemplate(startM, stopM));
				}
			} else if (dreamFirstStart) {
				Substitution.insertAll(processedClass,
						new DreamLifeCycleTemplateFirstStartAndStop(stopM,
								dreamFirstStartM, weaveable, processedClass.getActualClass()));
			} else {
				Substitution.insertAll(processedClass,
						new LifeCycleTemplateStop(stopM));
			}
		} else {
			if (start) {
				if (dreamFirstStart) {
					Substitution.insertAll(processedClass,
							new DreamLifeCycleTemplateFirstStartAndStart(
									startM, dreamFirstStartM, weaveable, processedClass.getActualClass()));
				} else {
					Substitution.insertAll(processedClass,
							new LifeCycleTemplateStart(startM));
				}
			} else if (dreamFirstStart) {
				Substitution.insertAll(processedClass,
						new DreamLifeCycleTemplateFirstStart(dreamFirstStartM,
								weaveable, processedClass.getActualClass()));
			}
		}

		// Manage prepareStopFc()

		if (dreamPrepareStop) {
			Substitution.insertAll(processedClass,
					new DreamLifeCyclePrepareStopTemplate(dreamPrepareStopM));
			if (!prepareStopLifeCycleControllerAlreadyImplemented(processedClass))
				processedClass.getSuperInterfaces().add(
						getFactory().Type().createReference(
								PrepareStopLifeCycleController.class));
		}

		if (start || stop || dreamFirstStart)
			if (!processedClass.getSuperInterfaces().contains(
					getFactory().Type().createReference(
							LifeCycleController.class))) {
				processedClass.getSuperInterfaces().add(
						getFactory().Type().createReference(
								LifeCycleController.class));
			}
	}

	private boolean prepareStopLifeCycleControllerAlreadyImplemented(
			CtClass<?> processedClass) {
		CtTypeReference<?> superClassDeclaration = processedClass.getReference();
		CtClass<?> superClass;
		while (superClassDeclaration != null) {
			superClass = (CtClass<?>) superClassDeclaration.getDeclaration();
			if (superClass.getSuperInterfaces().contains(
					getFactory().Type().createReference(
							PrepareStopLifeCycleController.class)))
				return true;
			superClassDeclaration = superClassDeclaration.getSuperclass();
		}
		return false;
	}

	private CtField<?> getServiceField(CtClass<?> target) {
		CtTypeReference<?> superClassDeclaration = target.getReference();
		CtClass<?> superClass;
		while (superClassDeclaration != null) {
			superClass = (CtClass<?>) superClassDeclaration.getDeclaration();
			CtField<?> serviceField = getServiceFieldInClass(superClass);
			if (serviceField != null) {
				return serviceField;
			}
			superClassDeclaration = superClassDeclaration.getSuperclass();
		}
		return null;
	}

	private CtField<?> getServiceFieldInClass(CtClass<?> clazz) {
		List<CtField<?>> fields = clazz.getFields();
		Service serviceAnnotation;
		if (DEBUG) System.out.print("Getting Service Field in " + clazz.getQualifiedName()
				+ " dream component \n");
		for (CtField<?> field : fields) {
			serviceAnnotation = field.getAnnotation(Service.class);
			if (serviceAnnotation != null) {
				if (DEBUG) System.out.print("Service Annotation found in "
						+ clazz.getQualifiedName() + " dream component \n");
				return field;
			}
		}
		return null;
	}

	/**
	 * Allows to add a new generator in the list
	 * 
	 * @param g
	 *            the new generator
	 */
	public void addGenerator(Generator g) {
		generators.add(g);
	}
}
